<?php

namespace Drupal\cforge\EventSubscriber;

use Drupal\mcapi\Entity\Storage\WalletStorage;
use Drupal\migrate\MigrateSkipRowException;
use Drupal\migrate\Event\MigratePreRowSaveEvent;
use Drupal\migrate\Event\MigratePostRowSaveEvent;
use Drupal\migrate\Event\MigrateImportEvent;
use Drupal\migrate\Row;
use Drupal\user\Entity\User;
use Drupal\user\Entity\Role;
use Drupal\field\Entity\FieldConfig;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Entity\Entity\EntityViewDisplay;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;


/**
 * Migration modifications
 *
 * Note the order of events in MigrateExecutable::import
 * Dispatch PRE_IMPORT
 *  $source->next
 *  $source->prepareRow
 *    hook_migrate_prepare_row
 *    hook_migrate_NAME_prepare_row
 *  while
 *    processRow
 *    Dispatch PRE_ROW_SAVE
 *    Dispatch POST_ROW_SAVE
 *  Dispatch POST_IMPORT
 *
 * @todo inject module_installer, account_switcher, module_handler, config_factory, entityTypemanager
 */
class MigrationSubscriber implements EventSubscriberInterface {


  /**
   * @todo move this to the migration definitions in the relevant modules
   */
  const FILTER_FORMATS = [
    'editor_full_html' => 'full_html',
    'editor_filtered_html' => 'basic_html',
    'filtered_html' => 'basic_html',
    'full_html' => 'full_html',
    'plain_text' => '', // This is the fallback
    'php_code' => '', // This will be ignored
    ''  => 'full_html'
  ];

  /**
   * @todo move this to the migration definitions in the relevant modules
   */
  const VOCABS = [
    'cforge_docs_categories' => 'binders',
    'offers_wants_types' => NULL,
    'offers_wants_categories' => 'category',
    'galleries' => 'galleries'
  ];

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return [
      'migrate.pre_import' => [['migratePreImport']],
      'migrate.pre_row_save' => [['migratePreRowSave']],
      'migrate.post_row_save' => [['migratePostRowSave']],
      'migrate.post_import' => [['migratePostImport']]
    ];
  }

  /**
   * @param Drupal\migrate\Event\MigrateImportEvent $event
   */
  public function migratePreImport(MigrateImportEvent $event) {
    $migration = $event->getMigration();
    switch ($migration->id()) {
      case 'd7_field_instance':
        // Find any duplicate field instances
        $db = $migration->getSourcePlugin()->getDatabase();
        $dupes = $db->query("SELECT min(id) FROM {field_config_instance} WHERE deleted = 0 GROUP BY entity_type, field_name, bundle HAVING count(id) > 1")->fetchCol();
        if ($dupes) {
          // Delete them from the ORIGINAL database
          $db->delete('field_config_instance')->condition('id', $dupes, 'IN')->execute();
        }
        break;
      case 'd7_user':
        // Prevent overwhelming the geopositioning service (module is re-enabled later)
        \Drupal::service('module_installer')->uninstall(['cforge_geo']);
        // Because the wallet names based on user names are wrong if user is not logged in.
        \Drupal::service('account_switcher')->switchTo(User::load(1));
        break;
      case 'd7_node_complete:forum':
        if (\Drupal::moduleHandler()->moduleExists('forum_access')) {
          \Drupal::service('module_installer')->install(['forum_access_migrate']);
        }
        break;
    }

  }

  /**
   * @param Drupal\migrate\Event\MigratePreRowSaveEvent $event
   */
  public function migratePreRowSave(MigratePreRowSaveEvent $event) {
    $row = $event->getRow();
    $migration = $event->getMigration();
    // Ensure the category->required setting is transferred.
    if ($migration->id() == 'd7_field_instance' and $row->getSourceProperty('field_name') == 'offers_wants_categories' and $row->getSourceProperty('entity_type') == 'transaction') {
      // Just take this one property which varied in some d7 sites then ditch the field
      $field = FieldConfig::load('mcapi_transaction.mcapi_transaction.category');
      $field->setRequired($row->getSourceProperty('required'))->save();
    }
    if ($migration->id() == 'd7_field' or $migration->id() == 'd7_field_instance') {
      //Skip these d7 fields and rename them when importing the entities they belong to.
      $existing_fields = [
        'manage_notes',
        'manage_responsibility',
        'profile_notes',
        'profile_phones',
        'profile_address',
        'profile_location',
        'upload', //became 'attached'
        'upload_private', // Became 'attached',
        'user_picture'
        //'title', // Prevents 'base_field_override' fields being created
      ];
      if (in_array($row->getSourceProperty('field_name'), $existing_fields)) {
        throw new MigrateSkipRowException('Field has already installed equivalent');
      }
      if ($row->getSourceProperty('field_name') == 'body' and $row->getSourceProperty('entity_type') == 'node') {
        throw new MigrateSkipRowException('Node body fields not migrated');
      }
      if ($row->getSourceProperty('field_name') == 'offers_wants_categories') {
        throw new MigrateSkipRowException('offers_wants_categories replacement is already installed by the smallads module');
      }
    }

    // All novel user fields should go on the 'profile' form display and 'default' display
    if ($migration->getPluginId() == 'd7_field_instance' and $row->getSourceProperty('entity_type') == 'user') {
      $formDisplay = EntityFormDisplay::load('user.user.profile');
      $formDisplay->setComponent($row->getSourceProperty('field_name'), ['weight' => 20])->save();
      $viewDisplay = EntityViewDisplay::load('user.user.default');
      $viewDisplay->setComponent($row->getSourceProperty('field_name'), ['weight' => 20])->save();
    }

    if ($migration->id() == 'd7_user') {
      if (!$row->getSourceProperty('status')) {
        if (!$this->hasTraded($migration->getSourcePlugin()->getDatabase(), $row->getSourceProperty('status'))) {
          throw new MigrateSkipRowException('Skipping blocked user');
        }
      }
      // Rename the fields
      $map = [
        'manage_notes' => 'notes_admin',
        'profile_notes' => 'notes',
        'profile_phones' => 'phones',
        'profile_address' => 'address',
        'profile_location' => 'coordinates',
      ];
      foreach ($map as $old_name => $new_name) {
        if ($row->hasDestinationProperty($old_name)) {
          $this->mapField($row, $old_name, $new_name);
        }
      }
    }

    if ($migration->id() == 'd7_mcapi_transaction') {
      $this->mapField($row, 'offers_wants_categories', 'category');
    }

    // Move all terms that used to be in offers_wants_categories field.
    if ($migration->getPluginId() == 'd7_node' or $migration->id() == 'd7_mcapi_transaction') {
      if ($row->hasDestinationProperty('offers_wants_categories')) {
        $entity_type = $row->getSourceProperty('entity_type');
        $new_field_name = $entity_type == 'node' ? 'categories' : 'category';
        $this->mapField($row, 'offers_wants_categories', $new_field_name);
      }
    }

    // Change the filter format for every body field
    // I don't know why this doesn't happen already
    if ($row->getSourceProperty('plugin') == 'd7_node') {
      $this->entityBodyFilterFormat($row, 'body');

      // Change the name of the image taxonomy field
      if ($migration->id() == 'd7_node:image') {
        $this->mapField($row, 'galleries', 'gallery');
      }
      // Change the name of the document taxonomy field
      if ($migration->id() == 'd7_node:document') {
        $this->mapField($row, 'cforge_docs_categories', 'doctype');
      }
      if ($row->hasSourceProperty('upload')) {
        $this->mapField($row, 'upload', 'attached');
      }
      if ($row->hasSourceProperty('upload_private')) {
        $this->mapField($row, 'upload_private', 'attached');
      }
    }

    if ($migration->id() == 'd7_custom_block') {
      // I don't know why this isn't happening automatically - perhaps a bug in d8.4?
      $row->setDestinationProperty('body/format', $this->convertFormat($row->getDestinationProperty('body/format'), 'full_html'));
    }

    // Comments
    if ($migration->id() == 'd7_comment_type') {
      if ($row->getSourceProperty('type') != 'forum') {
        throw new MigrateSkipRowException('Comments already exist');
      }
    }
    elseif ($migration->id() == 'd7_comment_field') {
      $row->setDestinationProperty('type', 'comments');
    }
    elseif ($migration->id() == 'd7_comment') {
      $this->entityBodyFilterFormat($row, 'comment_body');
      if ($row->getDestinationProperty('entity_type') == 'node') {
        if ($row->getSourceProperty('node_type') != 'forum') {
          print_r($row);
          $row->setDestinationProperty('comment_type', $row->getDestinationProperty('entity_type'));
          $row->setDestinationProperty('field_name', 'comments');
        }
      }
    }

    // Fix an error I can't explain
    elseif ($migration->id() == 'd7_field_formatter_settings') {
      if (empty($row->getDestinationProperty('view_mode'))) {
        throw new MigrateSkipRowException('Skipping because view mode is lost.');
      }
    }

    // Since we're not migrating the address field, we lose the country value in
    // its widget settings, so here we try to take the country from the d7 site
    // variable.
    //@todo ensure that d7 sites have the site_default_country variable populated..
    elseif ($migration->id() == 'd7_system_date') {
      // France as a default is a guess based on usage stags
      $default_country = $row->getSourceProperty('site_default_country');
      // @todo some sites don't have this var set so what to do?
      FieldConfig::load('user.user.address')
        ->setSetting('available_countries', [$default_country])
        ->setSetting('default_country', $default_country)
        ->save();
    }
    elseif ($migration->id() == 'd7_user_role') {
      $id = $row->getDestinationProperty('id');
      if ($id == 'anonymous' or $id == 'authenticated') {
        $permissions = $row->getDestinationProperty('permissions');
        $permissions[] = 'use text format basic_html';
        $permissions[] = 'use text format plain_text';
        $row->setDestinationProperty('permissions', array_unique($permissions));
      }
    }

  }

  /**
   * @param MigratePostRowSaveEvent $event
   * Save the user's avatar image which didn't migrate automatically
   */
  function migratePostRowSave(MigratePostRowSaveEvent $event) {
    if ($event->getMigration()->id() == 'd7_user') {
      $row = $event->getRow();
      if ($pic_id = $row->getSourceProperty('picture')) {
        User::load($row->getSourceProperty('uid'))
          ->set('user_picture', $pic_id)
          ->save();
        echo "\nPic for user $user_id: $pic_id";
      }
    }
  }

  /**
   * @param MigrateImportEvent $event
   */
  function migratePostImport(MigrateImportEvent $event) {
    switch ($event->getMigration()->id()) {
      case 'd7_user':
        //Install this only now to prevent mass geolocation on a free service.
        \Drupal::service('module_installer')->install(['cforge_geo']);
        break;
      case 'd7_filter_format':
        \Drupal::configFactory()->getEditable('filter.settings')->set('fallback_format', 'plain_text');
        break;
      case 'd7_mcapi_transaction':
        $this->removeSystemRole();
        break;
      case 'd7_node_complete:forum':
        \Drupal::service('module_installer')->uninstall(['forum_access_migrate']);
        break;
    }
  }

  /**
   * Delete all the accounts with the system role and delete the system role.
   */
  private function removeSystemRole() {
    $user_storage = \Drupal::entityTypeManager()->getStorage('user');
    $system_users = $user_storage->loadByProperties(['roles' => 'system']);
    $committee = $user_storage->loadByProperties(['roles' => 'committee']);
    $ladmins = $user_storage->loadByProperties(['roles' => 'local admin']);
    unset($system_users[1], $committee[1], $ladmins[1]);
    $ladmin = reset($ladmins) or $ladmin = array_shift($committee);
    foreach ($system_users as $account) {
      foreach (WalletStorage::walletsOf($account, TRUE) as $wallet) {
        //move it to local admin, with committee members having access
        $wallet->set('holder', $ladmin);
        if (\Drupal::moduleHandler()->moduleExists('mcapi_bursers')) {
          foreach ($committee as $comm) {
            $wallet->bursers->addBurser($comm);
          }
          $wallet->save();
        }
        echo "\nChanged owner and added bursers to former system wallet ".$wallet->label();
      }
      echo "\nDeleted former system user ".$account->getDisplayName();
      $account->delete();
    }
    if ($role = Role::load('system')){
      $role->delete();
      echo "\nDeleted system role";
    }
  }

  /**
   * Change the destination name of a field.
   * @param Row $row
   * @param string $old_name
   * @param string $new_name
   */
  private function mapField(Row $row, string $old_name, string $new_name) {
    $row->setDestinationProperty($new_name, $row->getDestinationProperty($old_name));
    $row->removeDestinationProperty($old_name);
  }

  /**
   * Convert the filter format of a body field
   * @param Row $row
   * @param string $body_field_name
   * @param type $default
   */
  private function entityBodyFilterFormat(Row $row, $body_field_name, $default = 'basic_html') {
    if ($bods = $row->getDestinationProperty($body_field_name)) {
      foreach ($bods as $delta => &$body) {
        $old = isset($body['format']) && $format = $body['format'];
        $body['format'] = $this->convertFormat($old, $default);
      }
      $row->setDestinationProperty($body_field_name, $bods);
    }
  }

  /**
   * Look up the new filter format witih the old
   */
  private function convertFormat($old_format_name, $default) {
    return static::FILTER_FORMATS[$old_format_name] ?? $default;
  }

  private function hasTraded($db, $uid) {
    return $db->query("SELECT count(xid) FROM mcapi_transactions WHERE payer = $uid or payee = $uid")->fetchField();
  }

}
